--[[
	The invoker executes commands in encounter data

	Terminology:

	A command line is every sequential pair of values (1,2), (3,4), (5,6), etc. in a command list
	A command list is an array of command lines
	A command bundle is an array of command lists

	Valid commands are:
		expect 				= {"<token or value> ... <token_n or value_n>","<op>","<token' or value'> ... <token_n' or value_n'>"}
		quash 				= "<alert>"
		set 					= {<var> = <token or value>, ..., <var_n> = <token_n or value_n> }
		alert 				= "<alert>"
		scheduletimer	   = {"<timer>",<token or number>}
		canceltimer 		= "<timer>"
		resettimer 			= [BOOLEAN]
		tracing 				= {<name>,...,<name_n>}
		unittracing		= {"boss1",...,"boss5"}
		proximitycheck 	= {"<token>",[10,11,18, or 28]}
		outproximitycheck	= {"<token>",[10,11,18, or 28]}
		raidicon 			= "<raidicon>"
		removeraidicon    = "<token>"
		arrow 				= "<arrow>"
		removearrow 		= "<token>"
		removeallarrows	= [BOOLEAN]
		invoke            = command bundle
		defeat            = [BOOLEAN]
		insert            = {"<userdata>", value},
		wipe              = "<userdata>"
		batchalert        = {"<alert>",...,"<alert_n>"}
		batchquash        = {"<alert>",...,"<alert_n>"}
		quashall          = [BOOLEAN]
		schedulealert     = {"<alert>",<token or number>}
		repeatalert       = {"<alert>",<token or number>}
		cancelalert       = "<alert>"
]]

local addon = DXE
local L = addon.L
local NID = addon.NID
local SN = addon.SN
local unit_to_unittarget = addon.Roster.unit_to_unittarget
local targetof = addon.targetof

local GetTime = GetTime
local wipe,concat = table.wipe,table.concat
local type,next,select = type,next,select
local ipairs,pairs,unpack = ipairs,pairs,unpack
local tostring,tonumber = tostring,tonumber
local band,bor = bit.band,bit.bor
local match,gmatch,gsub,find,split,sub,lower = string.match,string.gmatch,string.gsub,string.find,string.split,string.sub,string.lower
local UnitGUID, UnitName, UnitExists, UnitIsUnit = UnitGUID, UnitName, UnitExists, UnitIsUnit
local UnitBuff,UnitDebuff = UnitBuff,UnitDebuff

local pfl,key,CE,alerts,raidicons,arrows,announces,messages

local function RefreshProfile(db) pfl = db.profile end
addon:AddToRefreshProfile(RefreshProfile)

-- Temp variable environment
local userdata = {}

-- Command line handlers
local handlers = {}

---------------------------------------------
-- TABLE POOL
---------------------------------------------

local cache = {}
setmetatable(cache,{__mode = "k"})
local new = function()
	local t = next(cache)
	if t then
		cache[t] = nil
		return t
	else
		return {}
	end
end

local del = function(t)
	wipe(t)
	cache[t] = true
	return nil
end

---------------------------------------------
-- INITIALIZATION
---------------------------------------------

local module = addon:NewModule("Invoker","AceEvent-3.0","AceTimer-3.0")

addon.Invoker = module
local HW = addon.HW
local Alerts = addon.Alerts
local Arrows = addon.Arrows
local RaidIcons = addon.RaidIcons
local Radar = addon.Radar
-- command bundles executed upon events
local event_to_bundle = {}
local eventtype_to_bundle = {}
local combatbundle_to_filter = {}
local eventbundle_to_filter = {}

--[===[@debug@
local debug

local debugDefaults = {
	-- Related to function names
	Alerts = true,
	REG_EVENT = false,
	["handlers.set"] = false,
	replace_funcs = false,
	insert = false,
	tabinsert = true,
	tabupdate = true,
	wipe = false,
	wipe_container = false,
	["target.bossunit"] = false,
	["target.OBJECT_TARGET"] = false,
	["target.OBJECT_FOCUS"] = false,
	["target.StartPolling"] = false,
	["target.TeardownTarget"] = false,
	["target.failsafe"] = false,
	["target.scan"] = false,
	["target.try"] = false,
	["target.UNIT_TARGET"] = false,
	["target.fire"] = false,
}

--@end-debug@]===]

function module:OnInitialize()
	addon.RegisterCallback(self,"SetActiveEncounter","OnSet")
	addon.RegisterCallback(self,"StartEncounter","OnStart")
	addon.RegisterCallback(self,"StopEncounter","OnStop")
	
	--[===[@debug@
	self.db = addon.db:RegisterNamespace("Invoker", {
		global = {
			debug = debugDefaults
		},
	})

	debug = addon:CreateDebugger("Invoker",self.db.global,debugDefaults)
	--@end-debug@]===]
end

local class_color = {}
for class,color in pairs(RAID_CLASS_COLORS) do
	class_color[class] = ("|cff%02x%02x%02x"):format(color.r * 255, color.g * 255, color.b * 255)
end
-- If needed to go Global class color
local function getclasscolor(unit)
	if unit == nil then return "nil" end
	local classc = select(2,UnitClass(unit))
	if classc == nil and unit ~= nil then return unit end --- NPCs, BOSSs
	return class_color[classc]..unit.."|r" or unit
end
---------------------------------------------
-- EVENT TUPLES
---------------------------------------------
local tuple = {}

local function SetTuple(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14,a15,a16)
	tuple['1']  = a1  or "nil"
	tuple['2']  = a2  or "nil"
	tuple['3']  = a3  or "nil"
	tuple['4']  = a4  or "nil"
	tuple['5']  = a5  or "nil"
	tuple['6']  = a6  or "nil"
	tuple['7']  = a7  or "nil"
	tuple['8']  = a8  or "nil"
	tuple['9']  = a9  or "nil"
	tuple['10'] = a10 or "nil"
	tuple['11'] = a11 or "nil"
	tuple['12'] = a12 or "nil"
	tuple['13'] = a13 or "nil"
	tuple['14'] = a14 or "nil"
	tuple['15'] = a15 or "nil"
	tuple['16'] = a16 or "nil"
end

---------------------------------------------
-- CONTROLS
---------------------------------------------

function module:OnStart(_,...)
	if not CE then return end
	if next(eventtype_to_bundle) then
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED","COMBAT_EVENT")
	end
	for event in pairs(event_to_bundle) do
		self:RegisterEvent(event,"REG_EVENT")
	end
	addon:SetTracing(CE.onactivate.tracing,CE.name)
	-- Reset colors if not acquired
	for i,hw in ipairs(HW) do
		if hw:IsOpen() and not hw.tracer:First() then
			hw:SetInfoBundle("",1,1)
			hw:ApplyNeutralColor()
		end
	end
	if CE.onstart then
		self:InvokeCommands(CE.onstart,...)
	end
	if CE.enrage then
		self:Enrage(CE.enrage)
	end
	if CE.icon and pfl.Engage.EnableEngage then
		Alerts:Simple(CE.name..L.alert[" has been |cffC41F3BEngaged"],3,pfl.Engage.AudioEngage,pfl.Engage.MessageEngageColor,pfl.Engage.FlashScreenEngage,CE.icon,pfl.Engage.EngageTextSize)
	end
end

function module:OnStop()
	if not CE then return end
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	for event in pairs(event_to_bundle) do
		self:UnregisterEvent(event)
	end
	self:ResetUserData()
	Alerts:QuashByPattern("^invoker")
	Alerts:QuashByPattern("^raidauras")
	Alerts:QuashByPattern("^handler")
	Arrows:RemoveAll()
	RaidIcons:RemoveAll()
	--self:TeardownTarget()
	self:RemoveAllTimers()
	self:ResetAlertData()
	addon:HideRadar()
end

---------------------------------------------
-- ENRAGE HANDLER
---------------------------------------------
--[[
	Syntax in Encounters.lua:
	local data = {
		...
		enrage = {
			time10n = 600,
			time25n = 600,
			time10h = 600,
			time25h = 600,
		},
		...
	},
]]

do
	local diff_to_key = {
		[0] = "time25lfr",
		[1] = "time10n",
		[2] = "time25n",
		[3] = "time10h",
		[4] = "time25h",
		[5] = "timeflex", -- new Normal
		[6] = "timelfr",
		[7] = "timeheroic",
		[8] = "timemythic",
	}
	function module:DelayedEnrageMsg(time)
		local icon = "Interface\\Icons\\Spell_Shadow_UnholyFrenzy"
		if time == 60 then
			Alerts:NewTMessage(GetSpellInfo(8599).." ("..SecondsToTime(time)..")",pfl.Enrage.Message1mColor,icon,pfl.Enrage.Audio1m)
		elseif time == 30 then
			Alerts:NewTMessage(GetSpellInfo(8599).." ("..SecondsToTime(time)..")",pfl.Enrage.Message30Color,icon,pfl.Enrage.Audio30)
		elseif time == 15 then
			Alerts:NewTMessage(GetSpellInfo(8599).." ("..SecondsToTime(time)..")",pfl.Enrage.Message15Color,icon,pfl.Enrage.Audio15)	
		end
	end	
	function module:DelayedHalfEnrage(time)
		Alerts:NewTMessage(GetSpellInfo(8599).." ("..SecondsToTime(time)..")",pfl.Enrage.MessageHalfColor,"Interface\\Icons\\Spell_Shadow_UnholyFrenzy",pfl.Enrage.AudioHalf)			
	end		
	function module:DelayedFinalEnrage(time)
		Alerts:NewTMessage(GetSpellInfo(8599),pfl.Enrage.MessageEnrageColor,"Interface\\Icons\\Spell_Shadow_UnholyFrenzy",pfl.Enrage.AudioEnrage)			
	end		
	function module:Enrage(info)
		if not info then return end
		local diff = addon:GetRaidDifficulty()
		local key = diff_to_key[diff]

		local time = info[key] or nil
		local text = GetSpellInfo(8599)
		local icon = "Interface\\Icons\\Spell_Shadow_UnholyFrenzy"

		if time and time > 0 then
			if pfl.Enrage.Enabled then Alerts:Dropdown("invoker_enrage",text,time,pfl.Enrage.Flashtime,nil,pfl.Enrage.BarColor,pfl.Enrage.FlashBarColor,nil,icon,pfl.Enrage.AudioCd,pfl.Enrage.Audiotime) end
			if pfl.Enrage.EnableMessage then Alerts:NewTMessage(text.." ("..SecondsToTime(time)..")",pfl.Enrage.MessageColor,icon,pfl.Enrage.Audio)	end
			if pfl.Enrage.EnabledHalf then
				local half = time / 2
				local m = half % 60
				self:ScheduleTimer("DelayedHalfEnrage",half,half)
			end
			if pfl.Enrage.Enable1m then	self:ScheduleTimer("DelayedEnrageMsg",time - 60,60) end
			if pfl.Enrage.Enable30 then self:ScheduleTimer("DelayedEnrageMsg",time - 30,30) end
			if pfl.Enrage.Enable15 then self:ScheduleTimer("DelayedEnrageMsg",time - 15,15) end
			if pfl.Enrage.EnableEnrage then self:ScheduleTimer("DelayedFinalEnrage",time,time) end				
		end
	end
end


---------------------------------------------
-- REPLACES
---------------------------------------------

local ReplaceTokens

local function next_series_value(series,key)
	local ix,n = key.."_index",#series
	local i = userdata[ix]
	if i > n and not series.loop then
		i = n
	else
		i = ((i-1)%n)+1 -- Handles looping
		userdata[ix] = userdata[ix] + 1
	end
	return series[i]
end

local upvalue = ""

do
	local function tft()
		return HW[1].tracer:First() and HW[1].tracer:First().."target" or ""
	end
	-- targetchecks (code from BigWigs, thanks!)
	local t = nil
	local function buildTable()
		t = {
			"boss1", "boss2", "boss3", "boss4", "boss5",
			"target", "targettarget",
			"focus", "focustarget",
			"mouseover", "mouseovertarget",
			"party1target", "party2target", "party3target", "party4target",
			-- Faster
			"raid1target", "raid2target", "raid3target", "raid4target", "raid5target",
			"raid6target", "raid7target", "raid8target", "raid9target", "raid10target",
			"raid11target", "raid12target", "raid13target", "raid14target", "raid15target",
			"raid16target", "raid17target", "raid18target", "raid19target", "raid20target",
			"raid21target", "raid22target", "raid23target", "raid24target", "raid25target",
			"raid26target", "raid27target", "raid28target", "raid29target", "raid30target",
			"raid31target", "raid32target", "raid33target", "raid34target", "raid35target",
			"raid36target", "raid37target", "raid38target", "raid39target", "raid40target"
		}
		--for i = 1, 40 do t[#t+1] = string.format("raid%dtarget", i) end
		buildTable = nil
	end

	local function findTargetByGUID(id)
		local idType = type(id)
		
		if not t then buildTable() end
		for i, unit in next, t do
			if UnitExists(unit) and not UnitIsPlayer(unit) then
				local unitId = UnitGUID(unit)
				if idType == "number" then unitId = addon:ReturnUnitID(unitId) end
				if unitId == id then return unit end
			end
		end
	end

	local function GetUnitIdByGUID(mob)
		return findTargetByGUID(mob)
	end

	local function checkTarget(GUID)
		local mobId = GetUnitIdByGUID(GUID)
		--print("aaaaaaaaaaa",mobId)
		if mobId then
			local player = UnitName(mobId.."target")
			if player then return player else return 0 end
		end
	end
	function addon:CheckT(GUID)
		--print("ASDASD",checkTarget(GUID))
		checkTarget(GUID)
	end
	-- end targetchecks
 
	local RepFuncs = {
		--encdebug = function(text) return addon:EncDebug(text) end,
		playerguid = function() return addon.PGUID end,
		playername = function() return addon.PNAME end,
		vehicleguid  = function() return UnitGUID("vehicle") or "" end,
		vehiclenames = function() return concat(addon:VehicleNames(),", ") end,
		difficulty = function() return addon:GetRaidDifficulty() end,
		lfr = function() return addon:CheckLFR() end,
		flex = function() return addon:CheckFlex() end, -- WOD normal
		heroic = function() return addon:IsRaidHeroic() end,
		isheroic = function() return addon:IsHeroic() end, -- 5man heroic
		mythic = function() return addon:CheckMythic() end,
		maxplayers = function() return addon:GetMaxPlayers() end,
		throttle = function() return addon:Throttle(unit,delay) end,
		-- First health watcher
		tft = tft,
		tft_unitexists = function() return not not UnitExists(tft()) end,
		tft_isplayer = function() return not not UnitIsUnit(tft(),"player") end,
		tft_unitname = function(color)
			if not color then
				return UnitName(tft())
			else
				return getclasscolor(UnitName(tft()))
			end
		end,
		-- Role
		ishealer = function() return addon:IsHealer() end,
		istank = function(unit) return addon:IsTank(unit) end,
		isdps = function() return addon:IsDps() end,
		isranged = function() return addon:IsRanged() end,
		ismelee = function() return addon:IsMelee() end,
		iscaster = function() return addon:IsCaster() end,
		ismagicdispeller = function() return addon:IsMagicDispeller() end,
		istargettank = function(unit,boss) return addon:IsTargetTank(unit,boss) end,
		-- Utility
		dispell = function(dispelltype) return addon:Dispell(dispelltype) end,
		removeenrage = function() return addon:CanRemoveEnrage() end,
		interrupt = function(guid) return addon:CanInterrupt(guid) end,
		getworldstateinfo = function() 
			local text = select(4, GetWorldStateUIInfo(5))
			local time = tonumber(string.match(text or "", "%d+"))
			if not time then return 0 
			else return time end
		end,
		unitexist = function(unit)
			if UnitExists(unit) then return 1
			else return 0 end
		end,
		targetname = function(unit,unit2)
			local r
			if not unit2 then
				r = getclasscolor(UnitName(unit))
			else
				local enemyID = unit.."target"
				r = getclasscolor(UnitName(enemyID))
			end
			return r:gsub("%-.+", "").."|r|h"	
		end,
		bosstarget = function(unit)
			local target = unit.."target"
			local guid = UnitGUID(target)
			if not guid or addon:IsTargetTank(target,unit) or UnitDetailedThreatSituation(target, unit) ~= false or addon:ReturnUnitID(guid) ~= 1 then return false end
			return UnitName(target)
			--local r = getclasscolor(UnitName(target))
			--return r:gsub("%-.+", "").."|r|h"	
		end,
		bossname = function(unit)
			if UnitExists(unit) then
				return UnitName(unit)
			else
				return L.alert["Boss"]
			end  
		end,
		bossexist = function(id,bossguid)
			if id == nil then return false end
			for i= 1, 5 do
				local boss = ("boss%d"):format(i)
				if UnitExists(boss) == true then
					if tonumber(id) == addon:ReturnUnitID(UnitGUID(boss)) then
						if not bossguid then
							return true
						else
							return UnitGUID(boss)
						end
					end
				end
			end
			return false
		end,
		targetguild = function(onlyguid)
			local GUID = UnitGUID("target")
			if onlyguid then
				return GUID
			else
				return addon:ReturnUnitID(GUID)
			end
		end,
		focusguid = function(onlyguid)
			local GUID = UnitGUID("focus")
			if onlyguid then
				return GUID
			else
				return addon:ReturnUnitID(GUID)
			end
		end,		
		bossid = function(unit,onlyguid,unittarget)
			if unit ~= nil then
				local guid
				if unittarget then
					guid = UnitGUID(unit.."target")
				else
					guid = UnitGUID(unit)
				end
				if onlyguid then
					return guid
				else
					if guid ~= nil then
						return addon:ReturnUnitID(guid)
					else
						return 0
					end
				end
			end
		end,
		unittarget = function(unitID,returnguid,unitName)
			local guid, enemyID
			if not unitName then
				enemyID = unitID.."target"
				guid = UnitGUID(enemyID)
				if guid ~= nil then
					if returnguid then 
						return guid 
					else
						return addon:ReturnUnitID(guid)
					end
				else
					return 0
				end
			else -- "player"
				enemyID = "target"
				guid = UnitGUID(enemyID)
				if guid ~= nil then
					if returnguid then 
						return guid 
					else
						return addon:ReturnUnitID(guid)
					end
				else
					return 0
				end
			end
		end,
		targetid = function(guid) return GetUnitIdByGUID(guid) end,
		unitispet = function(unit) return addon:UnitIsPet(unit) end,
		unitisplayer = function(unit) return addon:UnitIsPlayer(unit) end,
		unitisdead = function(unit)
			if UnitIsDeadOrGhost(unit) == 1 then return 1
			else return 0 end
		end,
		colorplayer = function(player)
			if player then 
				local r = getclasscolor(player)
				return r:gsub("%-.+", "").."|r|h"
			end
		end,
		checkRaidDebuff = function(spellname,spellid,target)
			for i=1, GetNumGroupMembers() do
				local name = GetRaidRosterInfo(i)
				if UnitDebuff(name, spellname) then
					local _,_,_,_,_,_,_,_,_,_,id = UnitDebuff(name, spellname)
					if id == spellid and target ~= name then
						return name
					end
				end
			end
			return 0
		end,
		RaidDebuff = function(spellname)
			local raidnames = {}
			for i=1, GetNumGroupMembers() do
				local name = GetRaidRosterInfo(i)
				--if UnitAffectingCombat(name) and UnitDebuff(name, spellname) then
				if UnitDebuff(name, spellname) then
					raidnames[#raidnames+1] = name
				end
			end
			--print("raidnames",raidnames)
			return concat(raidnames," ")
		end,
		sync = function(prefix,msg,ext1,ext2)
			if prefix == nil then return "-" end
			if prefix == "BigWigs" then
				local bwPrefix, bwMsg = msg:match("^(%u-):(.+)")
				local sync, rest = msg:match("(%S+)%s*(.*)$")
				--local tmp = bwMsg:gsub("% .+", "")
				local tmp1,tmp2 = (" "):split(bwMsg)
				--print("BigWigs Test",tmp1,"-",tmp2)
				--print("BigWigs Test2",sync,"-",rest)
				if ext1 then
					return tmp2 or ""				
				else				
					return tmp1 or ""
				end
			elseif prefix == "DBM" then
				local DBMprefix, a, b, Dbmmsg, target, param3 = ("\t"):split(msg)
				--print("DBM Test",DBMprefix,a,b)
				if ext1 then
					--print("DBM Test3",target)
					return target or ""
				elseif ext2 then
					return param3 or ""
				else				
					--print("DBM Test2",Dbmmsg)
					return Dbmmsg
				end
			end
		end,
		getspellinfo  = function(spell) return GetSpellInfo(spell) end,
		ej_getspellinfo  = function(spell) return EJ_GetSectionInfo(spell) end,
		srcname_or_YOU = function()
			local r = getclasscolor(tuple['2'])
			return addon.PGUID == tuple['1'] and L.alert["YOU"] or r:gsub("%-.+", "").."|r|h"
		end,
		dstname_or_YOU = function()
			local r = getclasscolor(tuple['5'])
			return addon.PGUID == tuple['4'] and L.alert["YOU"] or r:gsub("%-.+", "").."|r|h"
		end,
		upvalue = function() return upvalue end,
		--- Functions with passable arguments
		-- Get's an alert's timeleft. Note: Add support if timeleft is ever used on a tagged alert
		timeleft = function(id,delta) return Alerts:GetTimeleft("invoker"..id) + (tonumber(delta) or 0) end,
		minustimeleft = function(id,delta) return Alerts:GetTimeleft("invoker"..id) - (tonumber(delta) or 0) end,
		
		gettime = function(delta,prev,length)
			if length then
				print("a",delta,prev,length)
				print("b",delta - (GetTime() - prev),GetTime()- prev)
				return math.floor(delta - (GetTime() - prev))
			elseif prev then
				return (tonumber(prev) or 0) - GetTime()
			elseif delta then
				return GetTime() + (tonumber(delta) or 0)
			else
				return GetTime()
			end
		end,
		getprev = function(prev)
			if prev then
				return GetTime() - (tonumber(prev) or 0)
			else
				return GetTime()
			end
		end,		
		
		npcid = function(guid) if guid then return NID[guid] else return "0" end end,
		targetdebuff = function(unit,debuff) return not not UnitDebuff(unit,debuff) end,
		targetdebuffdur = function(unit,debuff) local x = select(7,UnitDebuff(unit,debuff)) - GetTime() return format("%d",x) end,
		targetbuff = function(unit,buff) return not not UnitBuff(unit,buff) end,
		playerdebuff = function(debuff) return not not UnitDebuff("player",debuff) end,
		playerdebuffdur = function(debuff) local x = select(7,UnitDebuff("player",debuff)) - GetTime() return format("%d",x) end,
		playerbuff = function(buff) return not not UnitBuff("player",buff) end,
		playerbuffdur = function(buff) local x = select(7,UnitBuff("player",buff)) - GetTime() return format("%d",x) end,
		debuffstacks = function(unit,debuff) local c = select(4,UnitDebuff(unit,debuff)) return c end,
		buffstacks = function(unit,buff) local c = select(4,UnitBuff(unit,buff)) return c end,
		totalabsorb = function(unit)
			if unit then
				return UnitGetTotalAbsorbs("unit")* 100 or 0 
			end
		end,
		debuffabsorb = function(debuff, unit)
			local amount = select(15, UnitDebuff(unit, debuff))
			if amount then
				local maxhp = UnitHealthMax(unit)
				local pershield = math.max(1, math.floor(amount / maxhp * 100))
				if pershield <= 100 then
					return pershield
				else
					return 100
				end
			else
				return 0
			end
		end,
		
		hasicon = function(unit,icon) return RaidIcons:HasIcon(unit,icon) end,
		geticon = function(unit) return RaidIcons:GetIcon(unit) end,
		closest = function(container) return addon:FindClosestUnit(userdata[container]) end,
		HighestDistance = function(oplayer)
			SetMapToCurrentZone()
			local furthest, highestDistance = nil, 0
			for i=1, GetNumGroupMembers() do
				local unit = ("raid%d"):format(i)
				if UnitAffectingCombat(unit) and not UnitIsUnit(unit, oplayer) then
					local distance = addon:InRange(unit)
					if distance > highestDistance then
						highestDistance = distance
						furthest = unit
					end
				end
			end
			if furthest ~= nil then return UnitName(furthest) end
		end,
		inrange = function(unit) return addon:InRange(unit) end,
		unitposition = function(unit,x,y) local x1, y1 = UnitPosition(unit)	return x1 end,
		unitpositionold = function(unit,x,y)
			SetMapToCurrentZone()
			local x,y = GetPlayerMapPosition(unit)
			return x
		end,
		
		channeldur = function(unit) local name,sub,text,texture,start,finish = UnitChannelInfo(unit) if start and finish then return (finish - start) / 1000 else return 0 end end,
		castdur = function(unit) local name,sub,text,texture,start,finish = UnitChannelInfo(unit) if start and finish then return (finish - start) / 1000 else return 0 end end,
		unitcasting = function(unit) 
			local _, _, _, _, _, endTime = UnitCastingInfo(unit)
			local cast = endTime and (endTime / 1000 - GetTime()) or 0
			return cast or 0
		end,
		gethp = function(unit)
			if type(unit) == "number" then
				for i= 1, 5 do
					local boss = ("boss%d"):format(i)
					if UnitExists(boss) then
						if tonumber(unit) == addon:ReturnUnitID(UnitGUID(boss)) then
							local hp,hpmax = UnitHealth(boss),UnitHealthMax(boss) 
							if hp and hpmax ~= 0 then return hp / hpmax * 100 
							else return 100 end 					
						end
					end
				end
			elseif type(unit) == "string" then
				local hp,hpmax = UnitHealth(unit),UnitHealthMax(unit) 
				if hp and hpmax ~= 0 then return hp / hpmax * 100 
				else return 100 end 
			end		
		end,
		getup = function(unit,divide)
			if type(unit) == "number" then
				for i= 1, 5 do
					local boss = ("boss%d"):format(i)
					if UnitExists(boss) then
						if tonumber(unit) == addon:ReturnUnitID(UnitGUID(boss)) then
							local up,upmax = UnitPower(boss),UnitPowerMax(boss)
							if up and upmax ~= 0 then 
								if divide == nil then return up / upmax * 100
								else
									local d = up / upmax * 100
									return d / divide
								end
							else return 0 
							end 						
						end
					end
				end
			elseif type(unit) == "string" then
				local up,upmax = UnitPower(unit),UnitPowerMax(unit) 
				if up and upmax ~= 0 then
					if divide == nil then return up / upmax * 100
					else
						local d = up / upmax * 100
						return d / divide
					end
				else 
					return 0 
				end 
			end
		end,
		getap = function(unit) local ap = UnitPower(unit, ALTERNATE_POWER_INDEX) if ap then return ap else return 0 end end,
		getmultiraidap = function(value1,value2,radar)
			local unit, power 
			local players = {}
			--local players2 = {}
			for i=1, GetNumGroupMembers() do
				unit = GetRaidRosterInfo(i)
				local maxPower = UnitPowerMax(unit, ALTERNATE_POWER_INDEX)
				if maxPower ~= 0 and not UnitIsDeadOrGhost(unit) then
					power = UnitPower(unit, 10)
					if tostring(power) > tostring(value1) and tostring(power) < tostring(value2) then
						players[#players+1] = getclasscolor(UnitName(unit))
						--players2[#players2+1] = UnitName(unit)
					end
				end
			end
			--if not radar then
				local s ="" 
				for i,v in ipairs(players) do 
					s = s..v.." " 
				end 
				gsub(s,"%s$","") 
				return s
			--else
			--	return players2
			--end
		end,
		getraidap = function(value) 
			local unit, power
			local players = {}
			for i=1, GetNumGroupMembers() do
				unit = GetRaidRosterInfo(i)
				local maxPower = UnitPowerMax(unit, ALTERNATE_POWER_INDEX)
				if maxPower ~= 0 and not UnitIsDeadOrGhost(unit) then
					power = UnitPower(unit, 10)
					if tostring(power) == tostring(value) then
						players[#players+1] = getclasscolor(UnitName(unit))
					end
				end
			end
			local s ="" 
			for i,v in ipairs(players) do 
				s = s..v.." " 
			end 
			gsub(s,"%s$","") 
			return s
			--return players
		end,
		gettarget = function(guid) return checkTarget(guid) end,
		
		--container = function(container) local s ="" for i,v in ipairs(userdata[container]) do s = s..v.." " end gsub(s,"%s$","") return s end,	
		list = function(container) local s ="" for i,v in ipairs(userdata[container]) do s = s..v.." " end gsub(s,"%s$","") return s end,	
		contains = function(container,unit)
			--if tContains(userdata[container], unit) then return true end
			for i,v in ipairs(userdata[container]) do
				if v == unit then return true end
			end 
			return false
		end,
		count = function(container) 
			local s = 0 
			for i,v in ipairs(userdata[container]) do 
				s = s + 1 
			end 
			return s
		end,
		
		guidisplayertarget = function(guid) return UnitGUID("playertarget") == guid end,
		tabread = function(tab,key) local t = userdata[tab] return t[key] or 0 end,
		checkperc = function(var,value) 
			local v = var % value
			return v
		end,
		-- Encounters code
		checkPlayerSide = function() -- For SoO Spoils of Pandaria
			-- Thanks BigWigs for this code
			SetMapToCurrentZone()
			local cx, cy = GetPlayerMapPosition("player")
			if cy == 0 then return 0 end

			-- simplified cross product: above (mantid) > 0 (colinear) > below (mogu)
			return 0.04362700914 - (cx * 0.11017924547) + (cy * 0.04940152168)
		end,
		MargokJumpDistance = function(debuff,isFortification) -- WoD Highmaul Margok
			local _, _, _, amount = UnitDebuff("player", debuff)
			if not amount then amount = 1 end
			local jumpDistance = (isFortification and 0.75 or 0.5)^(amount - 1) * 200
			return jumpDistance
		end,
		checkPower = function(unit,newTime) -- For Blackrock Wod The Blast Furnance	
			local t = ceil((100-UnitPower(unit))/(100/newTime))
			return t
		end,
		
	}

	-- Add funcs for the other health watchers
	do
		for i=2,5 do
			local tft = function() return HW[i].tracer:First() and HW[i].tracer:First().."target" or "" end
			local tft_unitexists = function() return not not UnitExists(tft()) end
			local tft_isplayer = function() return not not UnitIsUnit(tft(),"player") end
			local tft_unitname = function() return UnitName(tft()) end
			RepFuncs["tft"..i] = tft
			RepFuncs["tft"..i.."_unitexists"] = tft_unitexists
			RepFuncs["tft"..i.."_isplayer"] = tft_isplayer
			RepFuncs["tft"..i.."_unitname"] = tft_unitname
		end
	end

	--[===[@debug@
	function module:GetRepFuncs()
		return RepFuncs
	end
	--@end-debug@]===]

	local replace_nums = tuple

	local function replace_nums2(num)
		if num == "2" then
			local r = getclasscolor(tuple['2'])
			return r:gsub("%-.+", "").."|r|h"
		elseif num == "5" then
			local r = getclasscolor(tuple['5'])
			return r:gsub("%-.+", "").."|r|h"
		else
			return tuple[num]
		end
	end

	local function replace_vars(str)
		local val = userdata[str]
		if type(val) == "table" then
			val = next_series_value(val,str)
		end
		return val
	end

	local function replace_funcs(str)
		if find(str,"|") then
			local func,args = match(str,"^([^|]+)|(.+)")
			func = RepFuncs[func]
			if not func then return end
			--if not args then return end
			--[===[@debug@
			debug("replace_funcs",format("func: %s ret: %s",str,tostring(func(split("|",args)))))
			--@end-debug@]===]
			--print("replace_funcs",format("func: %s ret: %s",str,tostring(func(split("|",args)))))
			return tostring(func(split("|",args)))
		else
			local func = RepFuncs[str]
			if not func then return end
			--[===[@debug@
			debug("replace_funcs",format("func: %s ret: %s",str,tostring(func())))
			--@end-debug@]===]
			return tostring(func())
		end
	end

	-- Replaces special tokens with values
	-- IMPORTANT: replace_funcs goes last
	function ReplaceTokens(str)
		if type(str) ~= "string" then return str end
		str = gsub(str,"#(.-)#",replace_nums)
		str = gsub(str,"<(.-)>",replace_vars)
		str = gsub(str,"&(.-)&",replace_funcs)
		return str
	end
	function ReplaceTokens2(str)
		if type(str) ~= "string" then return str end
		str = gsub(str,"<(.-)>",replace_vars)
		--print("ReplaceTokens2",str)
		str = gsub(str,"#(.-)#",replace_nums2)
		str = gsub(str,"&(.-)&",replace_funcs)
		return str
	end
end

---------------------------------------------
-- PREFILTER
-- for: alert, announce, arrow, raidicon
---------------------------------------------

local shortcuts = {
	srcself = function() return addon.PGUID == tuple['1'] end,
	srcother = function() return addon.PGUID ~= tuple['1'] end,
	dstself = function() return addon.PGUID == tuple['4'] end,
	dstother = function() return addon.PGUID ~= tuple['4'] end,
}

-- returns <var> or false
-- if false, then break execution in the command handler
local function prefilter(info,command)
	-- var is info
	if type(info) ~= "table" then return info end
	-- inlined expect
	local expect = info.expect
	if expect and not handlers.expect(expect) then
		return false
	end
	for k,cond in pairs(shortcuts) do
		if info[k] and cond() then
			handlers[command](info[k])
		end
	end
	-- var is the first index
	-- if nil, then break execution in the command handler
	return info[1]
end

---------------------------------------------
-- CONDITIONS
-- Credits to PitBull4's debug for this idea
---------------------------------------------

do
	local ops = {}

	ops['=='] = function(a, b) return a == b end
	ops['~='] = function(a, b) return a ~= b end
	--[[ops['=='] = function(a, b)
		if not a or not b then return false
		else return a == b end
	end
	ops['~='] = function(a, b)
		if not a or not b then return false
		else return a ~= b end
	end--]]
	ops['find'] = function(a,b) return find(a,b) end

	-- Intended to be used on numbers
	
	ops['>'] = function(a, b)
		a,b = tonumber(a),tonumber(b)
		if not a or not b then return false
		else return a > b end
	end

	ops['>='] = function(a, b)
		a,b = tonumber(a),tonumber(b)
		if not a or not b then return false
		else return a >= b end
	end

	ops['<'] = function(a, b)
		a,b = tonumber(a),tonumber(b)
		if not a or not b then return false
		else return a < b end
	end

	ops['<='] = function(a, b)
		a,b = tonumber(a),tonumber(b)
		if not a or not b then return false
		else return a <= b end
	end
	
	ops['%'] = function(a, b)
		a,b = tonumber(a),tonumber(b)
		if not a or not b then return false
		else return a % b end
	end
	
	local t = {}
	for k, v in pairs(ops) do t[#t+1] = k end
	for _, k in ipairs(t) do
		ops["not_" .. k] = function(a, b)
			return not ops[k](a, b)
		end
	end

	--[===[@debug@
	function module:GetConditions()
		return ops
	end
	--@end-debug@]===]

	local t = {}
	-- @ADD TO HANDLERS
	handlers.expect = function(info)
		if #info == 3 then
			return ops[info[2]](ReplaceTokens(info[1]),ReplaceTokens(info[3]))
		else
			-- there are at least two triplets
			--
			-- left to right association
			-- ex. (((a and b) or c) and d)
			--
			-- XXX A XXX A XXX A XXX
			--
			-- 3*x          + (x-1)           = 4x - 1 = total number
			-- ^					 ^
			-- num triplets    num logical ops

			local nres = (#info + 1) / 4
			for i=1,nres do
				-- left index of triplet
				local j = 4*i - 3
				local v1,op,v2 = info[j],info[j+1],info[j+2]
				--if not v1 or not v2 or not op then return false end
				local v = ops[op](ReplaceTokens(v1),ReplaceTokens(v2))
				t[i] = v
			end
			local ret = t[1]
			for i=2,nres do
				local ix = (i-1)*4
				local log_op = info[ix]
				if log_op == "AND" then
					ret = ret and t[i]
				elseif log_op == "OR" then
					ret = ret or t[i]
				end
			end
			return ret
		end
	end
end

---------------------------------------------
-- USERDATA
---------------------------------------------

do
	local wipeins = {} -- var -> handles

	local alert_keys = {
		"text",
		"time",
		"time10n",
		"time10h",
		"time25n",
		"time25h",
		"time25lfr",
		"timeflex", -- new Normal
		"timelfr",
		"timeheroic",
		"timemythic",
	}

	for i=2,15 do
		alert_keys[#alert_keys+1] = "text"..i
		alert_keys[#alert_keys+1] = "time"..i
	end

	function module:ResetUserData()
		wipe(userdata)
		for k,handle in pairs(wipeins) do
			module:CancelTimer(handle)
			wipeins[k] = nil
		end

		-- Copy alert time/text series into userdata
		-- time[2-9], text[2-9], time10n, time10h, time25n, time25h
		if CE.alerts then
			for var,info in pairs(CE.alerts) do
				for _,key in ipairs(alert_keys) do
					-- add userdata variable for this alert key
					local v = info[key]
					if type(v) == "table" then
						-- prepend '_' for safety
						-- ex. "_boltwarntime2"
						local ud_key = "_"..var..key
						userdata[ud_key] = v
						userdata[ud_key.."_index"] = 1
					end
				end
			end
		end

		if CE.userdata then
			-- Copy defaults into userdata
			for k,v in pairs(CE.userdata) do
				if type(v) == "table" then
					-- Indexing for series
					userdata[k.."_index"] = 1
					if v.type == "series" then
						userdata[k] = v
					elseif v.type == "container" then
						userdata[k] = {}
					end
				else
					userdata[k] = v
				end
			end
		end
	end

	-- @ADD TO HANDLERS
	handlers.set = function(info)
		for k,v in pairs(info) do
			local flag = true
			if type(v) == "string" then
				-- Increment/Decrement support
				v = ReplaceTokens(v)
				if find(v,"^INCR") then
					local delta = tonumber(match(v,"^INCR|(%d+)"))
					userdata[k] = userdata[k] + delta
					flag = false
				elseif find(v,"^DECR") then
					local delta = tonumber(match(v,"^DECR|(%d+)"))
					userdata[k] = userdata[k] - delta
					flag = false
				elseif find(v,"^MULT") then
					local delta = tonumber(match(v,"^MULT|(%d+)"))
					userdata[k] = userdata[k] * delta
					flag = false
				elseif find(v,"^PERC") then
					local delta = tonumber(match(v,"^PERC|(%d+)"))
					userdata[k] = userdata[k] % delta
					flag = false						
				--else
				--	v = ReplaceTokens(v)
				end
			end
			if flag then
				--[===[@debug@
				debug("handlers.set","var: <%s> before: %s after: %s",k,userdata[k],v)
				--@end-debug@]===]
				if type(v) == "table" then
					-- Indexing for series
					userdata[k.."_index"] = 1
					if v.type == "series" then
						userdata[k] = v
					elseif v.type == "container" then
						userdata[k] = {}
					end
				else
					userdata[k] = v
				end
			end
		end
		return true
	end

	local function wipe_container(k)
		wipeins[k] = nil
		wipe(userdata[k])
		--[===[@debug@
		debug("wipe_container","var: %s table values: %s",k,table.concat(userdata[k],", "))
		--@end-debug@]===]
	end

	-- @ADD TO HANDLERS
	handlers.insert = function(info)
		local k,v,n = info[1],info[2],info[3]
		--print("teste insert",k,v,t)
		if not n then
			v = ReplaceTokens2(v)
		else
			v = ReplaceTokens(v)
		end
		local t = userdata[k]
		t[#t+1] = v

		local ct = CE.userdata[k]
		if ct.wipein and not wipeins[k] then
			wipeins[k] = module:ScheduleTimer(wipe_container,ct.wipein,k)
		end

		--[===[@debug@
		debug("insert","var: %s value: %s table values: %s",k,v,table.concat(userdata[k],", "))
		--@end-debug@]===]

		return true
	end

	handlers.delete = function(info)
		local k,v,n = info[1],info[2],info[3]
		--print("teste delete",k,v,n)
		--if not n then
			local a = ReplaceTokens2(v)
		--else
			local b = ReplaceTokens(v)
		--end
		--local t = userdata[k]
		--t[#t+1] = v
		tDeleteItem(userdata[k], a)
		tDeleteItem(userdata[k], b)

		return true
	end		
	

	handlers.tabinsert = function(info)
		local t,k,v = info[1],info[2],info[3]
		k = ReplaceTokens(k)
		v = ReplaceTokens(v)
		local tab = userdata[t]
		if find(v,"^INCR") then
			local delta = tonumber(match(v,"^INCR|(%d+)"))
			tab[k] = tab[k] + delta
		elseif find(v,"^DECR") then
			local delta = tonumber(match(v,"^DECR|(%d+)"))
			tab[k] = tab[k] - delta
		else
			tab[k] = v
		end
		
		--print("tabinsert","tab:",t,"key:",k,"value:",v)
		--[===[@debug@
		debug("tabinsert","tab: %s key: %s value: %s",t,k,v)
		--@end-debug@]===]
	end

	handlers.tabupdate = function(info)
		local t,k,v = info[1],info[2],info[3]
		k = ReplaceTokens(k)
		local tab = userdata[t]
		--local oldv = tab[k]
		
		if tab[k] == nil then
			tab[k] = v 
			--print("tab[k] on tabupdate was nil",t,k,v)
		end

		if find(v,"^INCR") then
			local delta = tonumber(match(v,"^INCR|(%d+)"))
			tab[k] = tab[k] + delta
		elseif find(v,"^DECR") then
			local delta = tonumber(match(v,"^DECR|(%d+)"))
			tab[k] = tab[k] - delta
		elseif find(v,"^MULT") then
			local delta = tonumber(match(v,"^MULT|(%d+)"))
			tab[k] = tab[k] * delta			
		else
			v = ReplaceTokens(v)
			tab[k] = v
		end
		--[===[@debug@
		debug("tabupdate","tab: %s key: %s value: %s was: %s",t,k,v,oldv)
		--@end-debug@]===]

		return true
	end


	-- @ADD TO HANDLERS
	handlers.wipe = function(info)
		wipe(userdata[info])
		--[===[@debug@
		debug("wipe","var: %s table values: %s",info,table.concat(userdata[info],", "))
		--@end-debug@]===]
		return true
	end
end

---------------------------------------------
-- ALERTS
---------------------------------------------

do
	local throttles = {}
	local throttlespecial = {}
	local counters = {}

	function module:ResetAlertData()
		wipe(throttles)
		wipe(throttlespecial)
		wipe(counters)
	end

	local diff_to_key = {
		[0] = "time25lfr",
		[1] = "time10n",
		[2] = "time25n",
		[3] = "time10h",
		[4] = "time25h",
		[5] = "timeflex",
		[6] = "timemythic",
	}

	local function resolve_time(time,var,key)
		if type(time) == "table" then
			return next_series_value(time,"_"..var..key)
		elseif type(time) == "string" then
			return tonumber(ReplaceTokens(time))
		else
			return time
		end
	end

	local function resolve_text(text,var,key)
		if type(text) == "table" then
			return next_series_value(text,"_"..var..key)
		else
			return ReplaceTokens2(text)--ReplaceTokens(text,true) -- mop added, color nick
		end
	end

	local function getid(var)
		--if not alerts[var].tag then return end
		local tag = ReplaceTokens(alerts[var].tag) or ""
		tag = tag:gsub("%-.+", "")
		return "invoker"..var..tag
	end

	-- @ADD TO HANDLERS
	
	handlers.alert = function(info)
		local var = prefilter(info,"alert")
		if not var then return true end

		local stgs = pfl.Encounters[key][var]
		if not stgs then return true end
		local defn = alerts[var] or nil
		if not defn then return true end
		if stgs.encounterenabled == false and defn.enabled == true then
			stgs.encounterenabled = nil
			stgs.enabled = true
		end
		if stgs.enabled then
			if stgs.encounterenabled == nil and defn.enabled == false then
				stgs.enabled = false
				stgs.encounterenabled = false
				return true
			end

			-- Throttling
			if defn.throttle then
				-- Initialize to 0 if non-existant
				throttles[var] = throttles[var] or 0
				-- Check throttle
				local t = GetTime()
				--if GetTime() - throttles[var] > defn.throttle then
				if t - throttles[var] > defn.throttle then
				--if throttles[var] + defn.throttle < t then
					throttles[var] = t
				else
					-- Failed throttle, exit out
					return true
				end
			end

			if stgs.exlfr and addon:CheckLFR() then return true end
			if stgs.exflex and addon:CheckFlex() then return true end
			if stgs.exheroic and addon:IsRaidHeroic() then return true end
			if stgs.exmythic and addon:CheckMythic() then return true end

			if stgs.exhealer and addon:IsHealer() then return true end
			if stgs.extank and addon:IsTank() then return true end
			if stgs.exdps and addon:IsDps() then return true end
			if stgs.exmelee and addon:IsMelee() then return true end
			if stgs.exranged and addon:IsRanged() then return true end
			if stgs.excaster and addon:IsCaster() then return true end
			
			if defn.expect and not handlers.expect(defn.expect) then
				-- failed expect condition
				return true
			end

			local id = getid(var)
			Alerts:QuashByPattern(id)
			--[[
			local behavior = defn.behavior
			if behavior == "overwrite" then
				Alerts:QuashByPattern(id)
			elseif behavior == "singleton" then
				if Alerts:IsActive(id) then return true end
			end--]]

			local text,time
			-- Check to use specified text
			if info.text then
				local key = "text"..info.text
				local new_text = defn[key]
				text = resolve_text(new_text,var,key)
			end
			-- Replace text if it is still nil
			if not text then text = resolve_text(defn.text,var,"text") end
			
			-- counters
			if stgs.counter then
				local c = counters[var] or 0
				c = c + 1
				--text = text.." ("..c..")"
				text = "("..c..") "..text
				counters[var] = c
			end
			
			-- Time precedence
			-- 1. specified
			-- 2. difficulty
			-- 3. default

			-- Check to use specified time
			if info.time then
				local key = "time"..info.time
				local new_time = defn[key]
				time = resolve_time(new_time,var,key)
			end

			-- Check for difficulty time
			if not time then
				local diff = addon:GetRaidDifficulty()
				local key = diff_to_key[diff]
				local new_time = defn[key]
				time = new_time and resolve_time(new_time,var,key)
			end

			-- Replace time if it still nil
			if not time then
				time = resolve_time(defn.time,var,"time")
			end

			local flashtime = stgs.flashtime or defn.flashtime or pfl.Globals.TransitionTime 
			local audiocd = stgs.audiocd or defn.audiocd
			local sound = stgs.sound
			local flashscreen = stgs.flashscreen
			local icon = ReplaceTokens(info.icon) or ReplaceTokens(defn.icon)

			if defn.throttlespecial then
				-- Initialize to 0 if non-existant
				throttlespecial[var] = throttlespecial[var] or 0
				-- Check throttle
				local t = GetTime()
				if t - throttlespecial[var] > defn.throttlespecial then
					-- its fine
					throttlespecial[var] = t
				else
					-- Failed throttle, mute the sound/flash
					flashscreen = false
					audiocd = false
					sound = "None"
				end
			end
			
			-- Tag
			--local tag = ReplaceTokens(defn.tag) or ""

			--[===[@debug@
			debug("Alerts","id: %s var: %s text: %s time: %s flashtime: %s sound: %s color1: %s color2: %s tag: %s",id,var,text,time,defn.flashtime,stgs.sound,stgs.color1,stgs.color2,tag)
			--@end-debug@]===]
			-- Sanity check
			if not time or time < 0 then return true end
			
			local static = false
			if defn.static then static = true end
			
			-- Pass in appropriate arguments
			if stgs.type == "dropdown" then
				Alerts:Dropdown(id,text,time,flashtime,sound,stgs.color1,stgs.color2,flashscreen,icon,audiocd,stgs.audiotime,stgs.biggerbar,static)
			elseif stgs.type == "centerpopup" then
				Alerts:CenterPopup(id,text,time,flashtime,sound,stgs.color1,stgs.color2,flashscreen,icon,audiocd,stgs.audiotime,stgs.biggerbar,static)
			elseif stgs.type == "simple" then
				Alerts:Simple(text,time,sound,stgs.color1,flashscreen,defn.icon)
			elseif stgs.type == "debuff" then
				Alerts:DebuffPopup(id,text,time,flashtime,sound,stgs.color1,stgs.color2,flashscreen,icon,audiocd,stgs.audiotime,stgs.biggerbar,static)
			elseif stgs.type == "inform" then
				Alerts:Inform(text,sound,stgs.color1,time,flashscreen,icon)				
			elseif stgs.type == "message" then
				Alerts:NewTMessage(text,stgs.color1,icon,sound)				
			elseif stgs.type == "absorb" then
				Alerts:Absorb(id,text,defn.textformat,time,flashtime,sound,stgs.color1,stgs.color2,flashscreen,icon,
							  defn.values[tuple['7']],ReplaceTokens(defn.npcid))
			elseif stgs.type == "unitabsorb" then
				Alerts:UnitAbsorb(id,text,defn.textformat,time,flashtime,sound,stgs.color1,stgs.color2,flashscreen,icon,
							  defn.spellid,ReplaceTokens(defn.unit))							  
			end
		end
		return true
	end     

	handlers.batchalert = function(info)
		for _,info in ipairs(info) do handlers.alert(info) end
		return true
	end

	handlers.quash = function(info)
		Alerts:QuashByPattern(getid(info))
		--[===[@debug@
		debug("Alerts","QUASH: id: %s ",info)
		--@end-debug@]===]
		return true
	end

	handlers.quashpattern = function(info)
		local pattern = ReplaceTokens(info)
		Alerts:QuashByPattern(pattern)
		--[===[@debug@
		debug("Alerts","QUASHPATTERN: pattern: %s ",pattern)
		--@end-debug@]===]
		return true
	end

	handlers.batchquash = function(info)
		for _,var in ipairs(info) do handlers.quash(var) end
		return true
	end

	handlers.quashall = function(info)
		Alerts:QuashByPattern("^invoker")
		return true
	end

	handlers.settimeleft = function(info)
		local var,time = info[1],info[2]
		local id = getid(var)
		if type(time) == "string" then
			time = tonumber(ReplaceTokens(time))
		end
		if not time or time < 0 then return true end
		Alerts:SetTimeleft(id,time)
		return true
	end
	
	--- Sync Stuff
	handlers.sync = function(info)
		local var,opt = info[1],info[2]
		if opt == 1 or opt == nil then
			Alerts:SyncStuff(getid(var),var,1)
		else
			Alerts:SyncStuff(getid(var),var,2)
		end
		return true		
	end
	function module:Sync(id,var,opt)
		if opt == 1 or opt == nil then
			handlers.alert(var)
		else
			handlers.message(var)
		end
	end
	handlers.bigwigssync = function(info)
		local var = ReplaceTokens(info[1])
		local var2 = ReplaceTokens(info[2]) or ""
		local msg = strjoin(" ", var, var2)
		SendAddonMessage("BigWigs", "T:"..msg, IsPartyLFG() and "INSTANCE_CHAT" or "RAID")
	end
	handlers.dbmsync = function(info)
		--local DBMprefix, a, b, Dbmmsg = ("\t"):split(msg)
		local var = ReplaceTokens(info[1])
		local var2 = ReplaceTokens(info[2]) or ""
		local var3 = ReplaceTokens(info[3]) or ""
		local prefix = "M"
		local str = ("%s\t%s\t%s\t%s\t%s"):format(99999, 10680, var, var2, var3)
		SendAddonMessage("D4", prefix .. "\t" .. str, IsPartyLFG() and "INSTANCE_CHAT" or "RAID")
	end
	handlers.syncbosswhisper = function(info)
		local prefix, msg = info[1],info[2]
		SendAddonMessage(prefix, msg, IsInGroup(2) and "INSTANCE_CHAT" or "RAID")
	end
	
	handlers.message = function(info)
		local var = prefilter(info,"message")
		if not var then return true end

		local stgs = pfl.Encounters[key][var]
		if not stgs then return true end
		local defn = messages[var]
		if not defn then return true end
		if stgs.encounterenabled == false and defn.enabled == true then
			stgs.encounterenabled = nil
			stgs.enabled = true
		end
		if stgs.enabled then
			if stgs.encounterenabled == nil and defn.enabled == false then
				stgs.enabled = false
				stgs.encounterenabled = false
				return true
			end

			if defn.expect and not handlers.expect(defn.expect) then
				-- failed expect condition
				return true
			end

			-- Throttling
			if defn.throttle then
				-- Initialize to 0 if non-existant
				throttles[var] = throttles[var] or 0
				-- Check throttle
				local t = GetTime()
				if t - throttles[var] > defn.throttle then
					throttles[var] = t
				else
					-- Failed throttle, exit out
					return true
				end
			--[[else
				throttles[var] = throttles[var] or 0
				-- Check throttle
				local t = GetTime()
				if t - throttles[var] > 1 then
					throttles[var] = t
				else
					-- Failed throttle, exit out
					return true
				end	--]]		
			end
			
			if stgs.exlfr and addon:CheckLFR() then return true end
			if stgs.exflex and addon:CheckFlex() then return true end
			if stgs.exheroic and addon:IsRaidHeroic() then return true end
			if stgs.exmythic and addon:CheckMythic() then return true end

			if stgs.exhealer and addon:IsHealer() then return true end
			if stgs.extank and addon:IsTank() then return true end
			if stgs.exdps and addon:IsDps() then return true end
			if stgs.exmelee and addon:IsMelee() then return true end
			if stgs.exranged and addon:IsRanged() then return true end
			if stgs.excaster and addon:IsCaster() then return true end
			
			local text
			-- Check to use specified text
			if info.text then
				local key = "text"..info.text
				local new_text = defn[key]
				text = resolve_text(new_text,var,key)
			end
			-- Replace text if it is still nil
			if not text then text = resolve_text(defn.text,var,"text") end
			--text = text:gsub("%-.+", "")

			if stgs.counter then
				local c = counters[var] or 0
				c = c + 1
				text = text.." ("..c..")"
				counters[var] = c
			end
			local time = stgs.time or 2
			if defn.type == "message" then
				Alerts:NewTMessage(text,stgs.color1,defn.icon,stgs.sound)
			elseif defn.type == "inform" then
				Alerts:Inform(text,stgs.sound,stgs.color1,time,flashscreen,defn.icon)
			elseif defn.type == "simple" then
				Alerts:Simple(text,time,stgs.sound,stgs.color1,flashscreen,defn.icon)			
			end
		end
		return true
	end
	
---------------------------------------------
-- ANNOUNCES
---------------------------------------------
	local SendChatMessage = SendChatMessage

	-- @ADD TO HANDLERS
	handlers.announce = function(info)
		local var = prefilter(info,"announce")
		if not var then return true end
		local stgs = pfl.Encounters[key][var]
		local defn = announces[var] or nil
		if not defn then return true end

		if stgs.encounterenabled == false and defn.enabled == true then
			stgs.encounterenabled = nil
			stgs.enabled = true
		end
		if stgs.exlfr and addon:CheckLFR() then return true end
		if stgs.exflex and addon:CheckFlex() then return true end
		if stgs.exheroic and addon:IsRaidHeroic() then return true end
		if stgs.exmythic and addon:CheckMythic() then return true end
			
		if stgs.exhealer and addon:IsHealer() then return true end
		if stgs.extank and addon:IsTank() then return true end
		if stgs.exdps and addon:IsDps() then return true end
		if stgs.exmelee and addon:IsMelee() then return true end
		if stgs.exranged and addon:IsRanged() then return true end
		if stgs.excaster and addon:IsCaster() then return true end
		if stgs.enabled then
			--local defn = announces[var]
			if stgs.encounterenabled == nil and defn.enabled == false then
				stgs.enabled = false
				stgs.encounterenabled = false
				return true
			end
			local omsg = defn.msg
			--if stgs.ReplaceMe then
			if pfl.Announces.ReplaceMe then
				local pStart, pEnd, capt = find(lower(omsg), "(%s+)" .. lower(L.alert["Me"]));
				if pStart then
					omsg = format("%s%s%s%s", sub(omsg, 0, pStart - 1), capt, UnitName("player"), sub(omsg, pEnd + 1));		
				end
			end
			local msg = ReplaceTokens2(omsg)
			if stgs.type == "SAY" then
				SendChatMessage(msg,"SAY")
			elseif stgs.type == "YELL" then
				SendChatMessage(msg,"YELL")
			end
			
			-- Throttling
			-- Initialize to 0 if non-existant
			throttles[var] = throttles[var] or 0
			-- Check throttle
			local t = GetTime()
			--if GetTime() - throttles[var] > defn.throttle then
			local Throttle = stgs.throttlespecial or pfl.Announces.Throttle
			if t - throttles[var] > Throttle then
			--if throttles[var] + defn.throttle < t then
				throttles[var] = t
			else
				-- Failed throttle, exit out
				return true
			end
			if stgs.flashscreen and stgs.flashcolor then Alerts:FlashScreen(stgs.flashcolor) end
			if stgs.sound and not pfl.DisableSounds then Alerts:SoundT(stgs.sound) end
		end
		return true
	end

end

---------------------------------------------
-- SCHEDULING
---------------------------------------------

do
	local reg_timers = {}
	local alert_timers = {}
	local message_timers = {}

	function module:RemoveAllTimers()
		for name in pairs(reg_timers) do handlers.canceltimer(name) end
		for var in pairs(alert_timers) do handlers.cancelalert(var) end
		for var in pairs(message_timers) do handlers.cancelmessage(var) end
		-- Just to be safe
		self:CancelAllTimers()
	end

	local function schedule(info,timers,firefunc,cancelfunc,schedulefunc,store_args)
		local id,time = info[1],info[2]
		handlers[cancelfunc](id)
		timers[id] = new()

		if type(time) == "string" then
			time = tonumber(ReplaceTokens(time))
		end

		if not time or time < 0 then return true end

		if store_args then
			local args = new()
			-- Only need the first 7 (up to spellID)
			args[1],args[2],args[3],args[4],args[5],args[6],args[7] =
			tuple['1'],tuple['2'],tuple['3'],tuple['4'],tuple['5'],tuple['6'],tuple['7']

			timers[id].args = args
		end

		timers[id].handle = module[schedulefunc](module,firefunc,time,id)
	end

	local function cancel(info,timers)
		if timers[info] then
			module:CancelTimer(timers[info].handle,true)
			if timers[info].args then
				timers[info].args = del(timers[info].args)
			end
			timers[info] = del(timers[info])
		end
	end

	function module:FireTimer(name)
		-- Don't wipe reg_timers[name], it could be rescheduled
		self:InvokeCommands(CE.timers[name],unpack(reg_timers[name].args))
	end

	-- @ADD TO HANDLERS
	handlers.scheduletimer = function(info)
		schedule(info,reg_timers,"FireTimer","canceltimer","ScheduleTimer",true)
		return true
	end

	-- @ADD TO HANDLERS
	handlers.canceltimer = function(info)
		cancel(info,reg_timers)
		return true
	end

	function module:FireAlert(info)
		if alert_timers[info].args then
			SetTuple(unpack(alert_timers[info].args))
		end
		handlers.alert(info)
	end

	-- @ADD TO HANDLERS
	handlers.schedulealert = function(info)
		schedule(info,alert_timers,"FireAlert","cancelalert","ScheduleTimer",true)
		return true
	end

	-- @ADD TO HANDLERS
	handlers.repeatalert = function(info)
		schedule(info,alert_timers,"FireAlert","cancelalert","ScheduleRepeatingTimer",false)
		return true
	end

	-- @ADD TO HANDLERS
	handlers.cancelalert = function(info)
		cancel(info,alert_timers)
		return true
	end
	
	function module:FireMessage(info)
		if message_timers[info].args then
			SetTuple(unpack(message_timers[info].args))
		end
		handlers.message(info)
	end
	-- @ADD TO HANDLERS
	handlers.schedulemessage = function(info)
		schedule(info,message_timers,"FireMessage","cancelmessage","ScheduleTimer",true)
		return true
	end
	handlers.cancelmessage = function(info)
		cancel(info,message_timers)
		return true
	end
end

---------------------------------------------
-- ENGAGE TIMER
---------------------------------------------

do
	-- @ADD TO HANDLERS
	handlers.resettimer = function(info)
		addon:ResetTimer()
		return true
	end
end

---------------------------------------------
-- TRACING
---------------------------------------------

do
	-- @ADD TO HANDLERS
	handlers.tracing = function(info)
		addon:SetTracing(info,CE.name,CE.triggers.ejid or 0)
		return true
	end

	handlers.unittracing = function(info)
		addon:SetTracing(info,CE.name,CE.triggers.ejid or 0)
		return true
	end
	
	handlers.sortedtracing = function(info)
		addon:SetSortedTracing(info,CE.name,CE.triggers.ejid or 0)
		return true
	end	
	
	handlers.endcombat = function()
		addon:TriggerDefeat()
		return true
	end
	
	handlers.stopcombat = function()
		addon:TriggerEndCombat()
		return true
	end	
end

---------------------------------------------
-- RADAR
---------------------------------------------

do
	-- @ADD TO HANDLERS
	handlers.openwindow = function(info) 
		local range,players,multi,spellid = ReplaceTokens(info[1]),ReplaceTokens(info[2]),info[3],info[4]
		addon:OpenWindow(range,players,multi,spellid)
		return true
	end

	handlers.closewindow = function()
		addon:CloseWindow()
		return true
	end
end

---------------------------------------------
-- PROXIMITY CHECKING
---------------------------------------------

do
	local ProximityFuncs = addon:GetProximityFuncs()

	-- @ADD TO HANDLERS
	handlers.proximitycheck = function(info)
		local target,range = info[1],info[2]
		target = ReplaceTokens(target)
		return ProximityFuncs[range](target)
	end

	handlers.outproximitycheck = function(info)
		local target,range = info[1],info[2]
		target = ReplaceTokens(target)
		return not ProximityFuncs[range](target)
	end
end

---------------------------------------------
-- ARROWS
---------------------------------------------

do
	-- @ADD TO HANDLERS
	handlers.arrow = function(info)
		local var = prefilter(info,"arrow")
		if not var then return true end
		local defn = arrows[var]
		local stgs = pfl.Encounters[key][var]
		
		if stgs.encounterenabled == false and defn.enabled == true then
			stgs.encounterenabled = nil
			stgs.enabled = true
		end
		if stgs.enabled then
			if stgs.encounterenabled == nil and defn.enabled == false then
				stgs.enabled = false
				stgs.encounterenabled = false
				return true
			end
			local unit = ReplaceTokens(defn.unit)
			
			if stgs.exlfr and addon:CheckLFR() then return true end
			if stgs.exflex and addon:CheckFlex() then return true end
			if stgs.exheroic and addon:IsRaidHeroic() then return true end
			if stgs.exmythic and addon:CheckMythic() then return true end
			
			if stgs.exhealer and addon:IsHealer() then return true end
			if stgs.extank and addon:IsTank() then return true end
			if stgs.exdps and addon:IsDps() then return true end
			if stgs.exmelee and addon:IsMelee() then return true end
			if stgs.exranged and addon:IsRanged() then return true end
			if stgs.excaster and addon:IsCaster() then return true end

			if UnitExists(unit) then
				Arrows:AddTarget(unit,defn.persist,defn.action,defn.msg,defn.spell,stgs.sound,defn.fixed,
									 defn.xpos,defn.ypos,defn.range1,defn.range2,defn.range3,defn.lockon)
			end
		end
		return true
	end

	-- @ADD TO HANDLERS
	handlers.removearrow = function(info)
		info = ReplaceTokens(info)
		Arrows:RemoveTarget(info)
		return true
	end

	-- @ADD TO HANDLERS
	handlers.removeallarrows = function(info)
		Arrows:RemoveAll()
		return true
	end
end

---------------------------------------------
-- RAID ICONS
---------------------------------------------

do
	--[[
		 0 = no icon
		 1 = Yellow 4-point Star
		 2 = Orange Circle
		 3 = Purple Diamond
		 4 = Green Triangle
		 5 = White Crescent Moon
		 6 = Blue Square
		 7 = Red "X" Cross
		 8 = White Skull
	]]

	local function is_guid(str)
		--return type(str) == "string" and #str == 18 and str:find("%xx%x+")
		return type(str) == "string"
	end

	-- @ADD TO HANDLERS
	handlers.raidicon = function(info)
		local var = prefilter(info,"raidicon")
		if not var then return true end
		local stgs = pfl.Encounters[key][var]
		local defn = raidicons[var]
		
		if stgs.encounterenabled == false and defn.enabled == true then
			stgs.encounterenabled = nil
			stgs.enabled = true
		end
		if addon:IsPromoted() and stgs.enabled then
			if stgs.encounterenabled == nil and defn.enabled == false then
				stgs.enabled = false
				stgs.encounterenabled = false
				return true
			end
			if stgs.exlfr and addon:CheckLFR() then return true end	
			if stgs.exflex and addon:CheckFlex() then return true end
			if stgs.exheroic and addon:IsRaidHeroic() then return true end
			if stgs.exmythic and addon:CheckMythic() then return true end
			
			local unit = ReplaceTokens(defn.unit)
			if UnitExists(unit) then
				if defn.type == "FRIENDLY" then
					RaidIcons:MarkFriendly(unit,defn.icon,defn.persist)
				elseif defn.type == "MULTIFRIENDLY" then
					RaidIcons:MultiMarkFriendly(var,unit,defn.icon,defn.persist,defn.reset,defn.total)
				end
			elseif is_guid(unit) then
				if defn.type == "ENEMY" then
					RaidIcons:MarkEnemy(unit,defn.icon,defn.persist,defn.remove)
				elseif defn.type == "MULTIENEMY" then
					RaidIcons:MultiMarkEnemy(var,unit,defn.icon,defn.persist,defn.remove,defn.reset,defn.total)
				end
			elseif defn.id ~= nil then
				local starget = UnitGUID("target")
				local guid = addon:ReturnUnitID(starget)
				if guid == defn.id then
					if defn.type == "ENEMY" then
						RaidIcons:MarkEnemy(starget,defn.icon,defn.persist,defn.remove)
					elseif defn.type == "MULTIENEMY" then
						RaidIcons:MultiMarkEnemy(var,starget,defn.icon,defn.persist,defn.remove,defn.reset,defn.total)
					end
				end
			end
		end
		return true
	end

	-- @ADD TO HANDLERS
	handlers.removeraidicon = function(info)
		local unit = ReplaceTokens(info)
		if UnitExists(unit) then
			RaidIcons:RemoveIcon(unit)
		end
		return true
	end
	
	handlers.seticon = function(info)
		local unit = ReplaceTokens(info[1])
		local icon = ReplaceTokens(info[2]) or 0
		if UnitExists(unit) then
			SetRaidTarget(unit,icon)
		end
		return true
	end
end

---------------------------------------------
-- TARGET
---------------------------------------------

do

	--[[
	Target function
	How this will work:
	"target",{
		source = "#1#", -- this is mandatory
		wait = <time>, -- in seconds e.g. 0.2
		raidicon = <raidicon>,
		announce = <announce>,
		message = <message>,
		arrow = <arrow>,
		alerts = {
			self = <alert>,
			other = <alert>,
			unknown = <alert>,
		}
	}
	]]

	local info

	local function fire(unit)
		if UnitExists(unit) then
			upvalue,server = UnitName(unit)
			--local name = GetRaidRosterInfo(i)
		--	local teste
			if server then upvalue = upvalue.."-"..server end
			--print("fired1",upvalue,server)
			--upvalue = getclasscolor(UnitName(unit))
			if info.raidicon then
				handlers.raidicon(info.raidicon)
			end
			if info.arrow then
				--local range = addon:InRange(unit)
				--if server then teste = upvalue.."-"..server end
				--local range2 = addon:InRange(server)
				--print("arrow range",unit,range,info.checknearby)
				--handlers.arrow(info.arrow)
				--[[if info.arrowdef then
					if info.arrowdef == "<" and tonumber(range) < tonumber(info.arrowrange) then
						handlers.arrow(info.arrow)
					elseif info.arrowdef == ">" and tonumber(range) > tonumber(info.arrowrange) then
						handlers.arrow(info.arrow)
					end
				else--]]
					handlers.arrow(info.arrow)
				--end
			end
			if UnitIsUnit(upvalue,"player") then
				if info.announce then
					handlers.announce(info.announce)
				elseif info.alerts and info.alerts.self then
					handlers.alert(info.alerts.self)
				elseif info.alerts and info.alerts.batchself then
					handlers.batchalert(info.alerts.batchself)
				end				
			else
				if info.alerts then
					if info.alerts.other then
						upvalue = getclasscolor(UnitName(unit))
						handlers.alert(info.alerts.other)
					elseif info.alerts.batchother then
						upvalue = getclasscolor(UnitName(unit))
						handlers.batchalert(info.alerts.batchother)						
					elseif info.alerts.nearbymsg and info.checknearby then
						local range = addon:InRange(unit)
						--print("target range",range)
						if tonumber(range) < tonumber(info.checknearby) then
							handlers.alert(info.alerts.nearbymsg)
						end
					end
				end
			end
			upvalue = getclasscolor(UnitName(unit))
			--print("fired2",upvalue)
			if info.message then
				handlers.message(info.message)
			end
		end
	end

	local maxscans = 0
	local function checkt(sGUID)
		local t = nil
		local function buildTable()
			t = {
				"boss1", "boss2", "boss3", "boss4", "boss5",
				"target", "targettarget",
				"focus", "focustarget",
				"mouseover", "mouseovertarget",
				"party1target", "party2target", "party3target", "party4target",			
			}
			for i = 1, 40 do t[#t+1] = string.format("raid%dtarget", i) end
			--buildTable = nil
		end
		if not t then
			buildTable()
		end

		local targetname = nil
		for i, unit in next, t do
			if UnitGUID(unit) == sGUID then
				targetname, server = UnitName(unit.."target")
				if server then targetname = targetname.."-"..server end
				break
			end
		end
		--targetname, server = UnitName(sGUID.."target")
	--	if server then targetname = targetname.."-"..server end
		--print("target",targetname,server,sGUID)
		if targetname then
			if not addon:IsTargetTank(targetname) then
				if info.exclude then
					local _, class = UnitClass(targetname)
					if class == info.exclude then
						local range = addon:InRange(targetname)
						--print("Exclude1",class,targetname,range)
						if tonumber(range) <= tonumber(info.excludeyards) then
							--print("Exclude2")
							module:ScheduleTimer(checkt,info.wait,sGUID)
						else
							fire(targetname)
						end
					else
						fire(targetname)
					end
				else
					fire(targetname)
				end
			else
				--print("is tank, recheck",info.wait,sGUID,targetname)
				maxscans = maxscans + 1
				if tonumber(info.schedule) == maxscans then
					fire(targetname)
				else
					module:ScheduleTimer(checkt,info.wait,sGUID)
				end
			end
		elseif not targetname then
			--print("not found, recheck2",info.wait,sGUID,maxscans)
			if info.schedule then
				maxscans = maxscans + 1
				if info.schedule <= maxscans then
					module:ScheduleTimer(checkt,info.wait,sGUID)
				elseif maxscans > info.schedule then
					if info.alerts and info.alerts.unknown then
						handlers.alert(info.alerts.unknown)
					end				
				end
			else
				if info.alerts and info.alerts.unknown then
					handlers.alert(info.alerts.unknown)
				end		
				if info.alerts and info.alerts.unknownmsg then
					handlers.message(info.alerts.unknownmsg)
				end	
			end
		end
	end

	handlers.target = function(_info)
		info = _info
		maxscans = 0
		local sguid
		local sguid2
		if info.source then
			 sguid = ReplaceTokens(info.source)
		end
		if info.npcid then
			local bossid = tonumber(ReplaceTokens(info.npcid))
			for i = 1, 5 do
				local guid = UnitGUID("boss"..i)
				--local guidid = "boss"..i
				if guid then
					if addon:ReturnUnitID(guid) == tonumber(bossid) then
						sguid2 = guid
						--module:ScheduleTimer(checkt,info.wait,guid)
					end
				end
			end
		end
		if info.wait then
			if sguid2 then
				module:ScheduleTimer(checkt,info.wait,sguid2)
			else
				module:ScheduleTimer(checkt,info.wait,sguid)
			end
		else
			return true
		end
		return true
	end
end

do
	handlers.enrage = function(info)
		local text = select(1, GetSpellInfo(8599))
		local time = tonumber(ReplaceTokens(info[1]))
		local audiotime = tonumber(ReplaceTokens(info[2]))
		local icon = "Interface\\Icons\\Spell_Shadow_UnholyFrenzy"
		Alerts:QuashByPattern("handler_enrage")
		if audiotime then
			Alerts:Dropdown("handler_enrage",text,time,10,nil,"RED",nil,nil,icon,true,audiotime)
		else
			Alerts:Dropdown("handler_enrage",text,time,10,nil,"RED",nil,nil,icon)
		end
		return true
	end
end

--[[
do
	-- "target",{
	--		unit = <unit>			  -- OPTIONAL
	--		npcid = <npcid>,		  -- OPTIONAL
	-- 	raidicon = <raidicon>, -- fired when target exists
	--		announce = <announce>, -- fired when target is self 	  	-- condition: target exists
	--		arrow = <arrow>, 		  -- fired when target exists
	--		alerts = {
	--			self = <alert>,     -- fired when target is self 		-- condition: target exists
	--			other = <alert>,    -- fired when target is not self 	-- condition: target exists
	--			unknown = <alert>,  -- fired when target doesn't exist
	--		},
	-- }

	-- If a rogue is targeted by Defile and then vanishes then Lich King will recast
	-- it on a different person

	-- Scenario 1
	-- 1. npc triggers SPELL_CAST_START or SPELL_CAST_SUCCESS and srcFlags indicate that the npc is target or focus
	-- 2. Register for UNIT_TARGET. The following could happen:
	-- 	a. UNIT_TARGET is triggered for the boss within 0.3s so fire everything and teardown.
	--				OR
	--		b. UNIT_TARGET is not triggered within 0.3s because the npc did not change targets or focus/target was
	--			lost on the npc (unlikely to happen). If the npc is still a valid unit then fire everything for its
	--			target. Otherwise, scan	raidNtarget for the npc and fire everything for raidNtargettarget.

	-- Scenario 2
	-- 1. npc triggers SPELL_CAST_START or SPELL_CAST_SUCCESS and the npc is not a valid unit
	-- 2. Store npc's current target's guid. If there is no current target then fire info.alerts.unknown if it exists and stop execution.
	-- 3. Schedule a repeating timer every 0.05s	for a target change. Schedule a fail-safe for 0.3s later.
	-- 4. When a target changed is detected then fire everything.

	-- tuple['3'] is srcFlags
	-- Only call from a SPELL_CAST_START or SPELL_CAST_SUCCESS

	local FAILSAFE_TIME = 0.3
	local TRY_REPEAT_TIME = 0.05
	local MAX_TRIES = 6
	local OBJECT_TARGET = COMBATLOG_OBJECT_TARGET
	local OBJECT_FOCUS  = COMBATLOG_OBJECT_FOCUS
	local ut_unit			-- Assigned focus/target for UNIT_TARGET
	local info				-- Cached info
	local last_guid		-- GUID to check against for a target switch
	local tries				-- How many times a poll has been done to check for a target swich
	local try_handle		-- Handle for polling
	local cancel_handle	-- Handle for failsafe

	local function fire(unit)
		--[===[@debug@
		debug("target.fire","unit: %s UnitName: %s",unit,UnitName(unit or ""))
		--@end-debug@]===]
		if UnitExists(unit) then
			upvalue = UnitName(unit)
			if info.raidicon then
				handlers.raidicon(info.raidicon)
			end
			if info.arrow then
				handlers.arrow(info.arrow)
			end
			if UnitIsUnit(unit,"player") then
				if info.announce then
					handlers.announce(info.announce)
				end
				if info.alerts and info.alerts.self then
					handlers.alert(info.alerts.self)
				end
			else
				if info.alerts and info.alerts.other then
					handlers.alert(info.alerts.other)
				end
			end
		else
			if info.alerts and info.alerts.unknown then
				handlers.alert(info.alerts.unknown)
			end
		end
	end

	function module:UNIT_TARGET(_,unit)
		if unit == ut_unit then
			local npcid = NID[UnitGUID(unit)]
			--[===[@debug@
			debug("target.UNIT_TARGET","unit: %s npcid: %s",unit,npcid)
			--@end-debug@]===]
			if info.npcid == npcid then
				fire(targetof[unit])
				self:TeardownTarget()
			end
		end
	end

	local function scan(npcid)
		--[===[@debug@
		debug("target.scan","Invoked")
		--@end-debug@]===]
		for _,unit in pairs(unit_to_unittarget) do
				if	 UnitExists(unit)
				and NID[UnitGUID(unit)] == npcid
				and UnitExists(targetof[unit]) then
				return targetof[unit]
			end
		end
	end

	local function try()
		tries = tries + 1
		local unit = scan(info.npcid)
		local cancel

		--[===[@debug@
		debug("target.try","tries: %s UnitName: %s",tries,UnitName(unit or ""))
		--@end-debug@]===]

		-- target changed
		if unit and UnitGUID(unit) ~= last_guid then
			fire(unit)
			cancel = true
		end

		if cancel then
			module:TeardownTarget()
		elseif tries == MAX_TRIES then
			-- failsafe fire on current target
			fire(unit or "")
			module:TeardownTarget()
		end
	end

	local function failsafe()
		--[===[@debug@
		debug("target.failsafe","Invoked")
		--@end-debug@]===]
		fire(targetof[ut_unit])
		cancel_handle = nil
		module:TeardownTarget()
	end

	function module:TeardownTarget()
		if try_handle then
			self:CancelTimer(try_handle)
			try_handle = nil
		end
		if cancel_handle then
			self:CancelTimer(cancel_handle)
			cancel_handle = nil
		end
		self:UnregisterEvent("UNIT_TARGET")
		--[===[@debug@
		debug("target.TeardownTarget","Invoked")
		--@end-debug@]===]
	end

	-- @ADD TO HANDLERS
	handlers.target = function(_info)
		info = _info
		module:TeardownTarget()
		if info.unit then
			module:RegisterEvent("UNIT_TARGET")
			ut_unit = info.unit
			cancel_handle = module:ScheduleTimer(failsafe,FAILSAFE_TIME)
			--[===[@debug@
			debug("target.bossunit","UnitName: %s",UnitName(targetof[info.unit]))
			--@end-debug@]===]
		elseif band(tuple['3'],OBJECT_TARGET) == OBJECT_TARGET then
			module:RegisterEvent("UNIT_TARGET")
			ut_unit = "target"
			cancel_handle = module:ScheduleTimer(failsafe,FAILSAFE_TIME)
			--[===[@debug@
			debug("target.OBJECT_TARGET","UnitName: %s",UnitName("targettarget"))
			--@end-debug@]===]
		elseif band(tuple['3'],OBJECT_FOCUS) == OBJECT_FOCUS then
			module:RegisterEvent("UNIT_TARGET")
			ut_unit = "focus"
			cancel_handle = module:ScheduleTimer(failsafe,FAILSAFE_TIME)
			--[===[@debug@
			debug("target.OBJECT_FOCUS","UnitName: %s",UnitName("focustarget"))
			--@end-debug@]===]
		else
			local unit = scan(info.npcid)
			if unit then last_guid = UnitGUID(unit)
			else last_guid = nil end
			try_handle = module:ScheduleRepeatingTimer(try,TRY_REPEAT_TIME)
			tries = 0
			--[===[@debug@
			debug("target.StartPolling","UnitName: %s last_guid: %s",UnitName(unit or ""),last_guid)
			--@end-debug@]===]
		end
		return true
	end
end]]

---------------------------------------------
-- INVOKING
---------------------------------------------

do
	local flag = true

	-- @param bundle Command bundles
	-- @param ... arguments passed with the event
	function module:InvokeCommands(bundle,...)
		if flag then SetTuple(...) end
		for _,list in ipairs(bundle) do
			for i=1,#list,2 do
				local type,info = list[i],list[i+1]
				local handler = handlers[type]
				-- Make sure handler exists in case of an unsupported command
				if handler and not handler(info) then break end
			end
		end
	end

	-- @ADD TO HANDLERS
	handlers.invoke = function(info)
		-- tuple has already been set
		flag = false; module:InvokeCommands(info); flag = true
		return true
	end
end

---------------------------------------------
-- EVENTS
---------------------------------------------

do
	local throttles = {}  -- bundle -> throttle time
	local times = {}      -- bundle -> last time

	local function main_event_handler(bundles,bundle_to_filter,attr_handles,...)
		if not bundles then return end
		for _,bundle in ipairs(bundles) do
			local skip = false
			if throttles[bundle] then
				times[bundle] = times[bundle] or 0
				local t = GetTime()
				if times[bundle] + throttles[bundle] < t then
					times[bundle] = t
				else
					skip = true
				end
			end
			if not skip then
				local filter = bundle_to_filter[bundle]
				local flag = true
				for attr,data in pairs(filter) do
					-- all conditions have to pass for the bundle to fire
					if not attr_handles[attr](data,...) then
						flag = false
						break
					end
				end
				if flag then
					module:InvokeCommands(bundle,...)
				end
			end
		end
	end

	local TYPE_PLAYER = COMBATLOG_OBJECT_TYPE_PLAYER
	local TYPE_NPC = COMBATLOG_OBJECT_TYPE_NPC

	local combat_attr_handles = {
		spellid = function(hash,...)
			local spellid = select(7,...)
			return hash[spellid]
		end,

		spellid2 = function(hash,...)
			local spellid2 = select(10,...)
			return hash[spellid2]
		end,

		spellname = function(hash,...)
			local spellname = select(8,...)
			return hash[spellname]
		end,

		spellname2 = function(hash,...)
			local spellname2 = select(11,...)
			return hash[spellname2]
		end,

		srcnpcid = function(hash,...)
			local srcguid = ...
			return hash[NID[srcguid]]
		end,

		dstnpcid = function(hash,...)
			local dstguid = select(4,...)
			return hash[NID[dstguid]]
		end,

		srcisplayertype = function(bool,...)
			local _,_,srcflags = ...
			return (band(srcflags,TYPE_PLAYER) == TYPE_PLAYER) == bool
		end,

		srcisnpctype = function(bool,...)
			local _,_,srcflags = ...
			return (band(srcflags,TYPE_NPC) == TYPE_NPC) == bool
		end,

		srcisplayerunit = function(bool,...)
			local srcguid = ...
			return (srcguid == addon.PGUID) == bool
		end,

		dstisplayertype = function(bool,...)
			local dstflags = select(6,...)
			return (band(dstflags,TYPE_PLAYER) == TYPE_PLAYER) == bool
		end,

		dstisnpctype = function(bool,...)
			local dstflags = select(6,...)
			return (band(dstflags,TYPE_NPC) == TYPE_NPC) == bool
		end,

		dstisplayerunit = function(bool,...)
			local dstguid = select(4,...)
			return (dstguid == addon.PGUID) == bool
		end,
	}

	local combat_transforms = {
		spellname = function(spellid) return SN[spellid] end,
		spellname2 = function(spellid) return SN[spellid] end,
	}

	--local weare42 = tonumber((select(4, GetBuildInfo()))) > 40100
	--if weare42 then
		function module:COMBAT_EVENT(event,timestamp,eventtype,hideCaster,srcGUID, srcName, srcFlags, srcFlags2, dstGUID, dstName, dstFlags, dstFlags2, ...)
			main_event_handler(eventtype_to_bundle[eventtype],combatbundle_to_filter,combat_attr_handles, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
		end
	--else
	--	function module:COMBAT_EVENT(event,timestamp,eventtype,hideCaster,...)
	--		main_event_handler(eventtype_to_bundle[eventtype],combatbundle_to_filter,combat_attr_handles,...)
	--	end
	--end

	local REG_ALIASES = {
		YELL = "CHAT_MSG_MONSTER_YELL",
		EMOTE = "CHAT_MSG_RAID_BOSS_EMOTE",
		MONSTER_EMOTE = "CHAT_MSG_MONSTER_EMOTE",
		BOSS_EMOTE = "RAID_BOSS_EMOTE",
		WHISPER = "CHAT_MSG_RAID_BOSS_WHISPER",
		BOSS_WHISPER = "RAID_BOSS_WHISPER",
	}

	local reg_attr_handles = {
		msg = function(hash,...)
			local msg = ...
			for str in pairs(hash) do
				if find(msg,str) then
					return true
				end
			end
		end,

		spellname = function(hash,...)
			local _,spell = ...
			return hash[spell]
		end,

		npcname = function(hash,...)
			local _,npcname = ...
			return hash[npcname]
		end,
	}

	local reg_transforms = {
		spellname = function(spellid) return SN[spellid] end,
	}

	function module:REG_EVENT(event,...)
		main_event_handler(event_to_bundle[event],eventbundle_to_filter,reg_attr_handles,...)
	end

	local function to_hash(work,trans)
		if type(work) ~= "table" then
			if trans then work = trans(work) end
			return {[work] = true}
		else
			-- work is an array
			local t = {}
			for k,v in pairs(work) do
				if trans then v = trans(v) end
				t[v] = true
			end
			return t
		end
	end

	-- filter structure = {
	-- 	[attribute] = {
	-- 		[value] = true,
	--			...
	--			[valueN] = true,
	--		},
	--    .
	--    .
	--    .
	-- 	[attributeN] = {
	-- 		[value] = true,
	--			...
	--			[valueN] = true,
	--		},
	-- }


	local function create_filter(info,handles,transforms)
		local filter = {}
		for attr in pairs(handles) do
			if info[attr] then
				if type(info[attr]) == "boolean" then
					filter[attr] = info[attr]
				else
					filter[attr] = to_hash(info[attr],transforms[attr])
				end
			end
		end
		return filter
	end

	function module:AddEventData()
		if not CE.events then return end
		-- Iterate over events table
		for _,info in ipairs(CE.events) do
			if info.type == "combatevent" then
				local t = eventtype_to_bundle[info.eventtype]
				if not t then
					t = {}
					eventtype_to_bundle[info.eventtype] = t
				end
				t[#t+1] = info.execute

				combatbundle_to_filter[info.execute] = create_filter(info,combat_attr_handles,combat_transforms)
			elseif info.type == "event" then
				local event = REG_ALIASES[info.event] or info.event
				local t = event_to_bundle[event]
				if not t then
					t = {}
					event_to_bundle[event] = t
				end
				t[#t+1] = info.execute

				eventbundle_to_filter[info.execute] = create_filter(info,reg_attr_handles,reg_transforms)
			end
			if info.throttle then throttles[info.execute] = info.throttle end
		end
	end

	function module:WipeEvents()
		wipe(event_to_bundle)
		wipe(eventtype_to_bundle)
		wipe(combatbundle_to_filter)
		wipe(eventbundle_to_filter)
		wipe(throttles)
		wipe(times)
		self:UnregisterAllEvents()
	end
end


---------------------------------------------
-- TRACER ACQUIRES
---------------------------------------------
-- Holds command bundles
local AcquiredBundles = {}
local UnitIsDead = UnitIsDead

function module:HW_TRACER_ACQUIRED(_,unit,npcid)
	if AcquiredBundles[npcid] and not UnitIsDead(unit) then
		self:InvokeCommands(AcquiredBundles[npcid])
	end
end

-- Each entry in
function module:SetOnAcquired()
	wipe(AcquiredBundles)
	local onacquired = CE.onacquired
	if not onacquired then return end
	for npcid,bundle in pairs(onacquired) do
		AcquiredBundles[npcid] = bundle
	end
end

addon.RegisterCallback(module,"HW_TRACER_ACQUIRED")

-- Debug Stuff
handlers.debug = function(info)
	--info = ReplaceTokens(info)
	addon:EncDebug(ReplaceTokens(info[1]),ReplaceTokens(info[2]),ReplaceTokens(info[3]),ReplaceTokens(info[4]),ReplaceTokens(info[5]))
	return true
end

---------------------------------------------
-- API
---------------------------------------------

function module:OnSet(_,data)
	--[===[@debug@
	assert(type(data) == "table","Expected 'data' table as argument #1 in OnSet. Got '"..tostring(data).."'")
	--@end-debug@]===]
	-- Set upvalues
	CE = data
	arrows = CE.arrows
	raidicons = CE.raidicons
	alerts = CE.alerts
	messages = CE.messages
	announces = CE.announces
	key = CE.key
	-- Wipe events
	self:WipeEvents()
	-- Register events
	self:AddEventData()
	-- Copy data.userdata to userdata upvalue
	self:ResetUserData()
	-- OnAcquired
	self:SetOnAcquired()
end
function module:ReplaceTokens(info)
	if info then return ReplaceTokens(info) end
end